"""
Module for bot related functionality.

This module implements the `Bot` class that manage the communication between
the user (through a telegram chat) and the camera.
"""
import inspect
import logging
import os
import sys
from functools import wraps
from typing import Any, Callable, Optional, Union

from telegram import ChatAction, ParseMode, ReplyKeyboardMarkup, Update
from telegram.ext import (
    CallbackContext,
    CommandHandler,
    Dispatcher,
    PicklePersistence,
    Updater,
    run_async
)

from surveillance_bot.bot_config import BotConfig
from surveillance_bot.camera import (
    Camera,
    CameraConnectionError,
    CodecNotAvailable
)

HandlerType = Callable[[Update, CallbackContext], Any]


class Bot:
    """
    Class for the telegram bot implementation.

    This class exposes a number of commands to the user in order to control
    the camera processes and receive picture and video files.

    Args:
        token: Access Token for the telegram bot.
        username: Username of the only user authorized to interact with the
            bot (without @).
        log_level: Logging level for logging module.
    """
    def __init__(
            self,
            token: str,
            username: str,
            persistence_dir: Optional[str] = None,
            log_level: Union[int, str, None] = None
    ) -> None:
        self.logger = logging.getLogger(__name__)
        if log_level:
            self.logger.setLevel(log_level)

        if not token:
            self.logger.critical("Error! Missing BOT_API_KEY configuration")
            sys.exit(1)
        if not username:
            self.logger.critical(
                "Error! Missing AUTHORIZED_USER configuration"
            )
            sys.exit(1)

        try:
            self.camera = Camera()
        except CameraConnectionError:
            self.logger.critical("Error! Can not connect to the camera.")
            sys.exit(2)
        except CodecNotAvailable:
            self.logger.critical(
                "Error! There are no suitable video codec available."
            )
            sys.exit(2)

        self.authorized_user = username

        persistence: Optional[PicklePersistence]
        if persistence_dir:
            os.makedirs(persistence_dir)
            path = os.path.join(persistence_dir, 'surveillance-bot.pickle')
            persistence = PicklePersistence(filename=path)
        else:
            persistence = None

        self.updater = Updater(
            token=token,
            persistence=persistence,
            use_context=True
        )

        dispatcher: Dispatcher = self.updater.dispatcher

        # Registers commands in the dispatcher
        for name, method in inspect.getmembers(self, inspect.ismethod):
            if name.startswith('_command_'):
                command = name.replace('_command_', '')
                dispatcher.add_handler(self.command_handler(command, method))

        # Registers configuration menu
        dispatcher.add_handler(BotConfig.get_config_handler(self))

        # Register error handler
        dispatcher.add_error_handler(self._error)

    def command_handler(
            self,
            command: str,
            callback: HandlerType
    ) -> CommandHandler:
        """
        Decorates callback and returns a CommandHandler.

        This decorator restricts command use to the authorized user, loads
        defaults configuration options and adds debug logging.

        Args:
            command: The command this handler should listen for.
            callback: The callback function for this handler.

        Returns:
            Handler instance to handle Telegram commands.
        """
        logger = self.logger

        @wraps(callback)
        def wrapped(update: Update, context: CallbackContext) -> Any:

            # Checks if user is authorized
            if update.effective_chat.username != self.authorized_user:
                logger.warning(
                    'Unauthorized call to "%s" command by @%s',
                    command,
                    update.effective_chat.username
                )
                update.message.reply_text(text="Unauthorized")
                return None

            BotConfig.ensure_defaults(context)
            logger.debug('Received "%s" command', command)
            return callback(update, context)

        return CommandHandler(command, wrapped)

    def start(self) -> None:
        """
        Starts the bot execution and waits to clean up before exit.

        After starting the camera and the bot polling it waits into a loop
        until the bot is interrupted by a signal. After that the camera
        device is released and the function ends.
        """
        self.camera.start()
        self.updater.start_polling()
        self.logger.info("Surveillance Bot started")

        self.updater.idle()

        self.camera.stop()
        self.logger.info("Surveillance Bot stopped")

    def _error(self, update: Update, context: CallbackContext) -> None:
        """
        Logs Errors caused by updates.

        Args:
            update: The update to be handled.
            context: The context object for the update.
        """
        self.logger.warning(
            'Update "%s" caused error "%s"',
            update,
            context.error
        )
        context.bot.send_message(
            chat_id=update.effective_chat.id,
            text="*ERROR|!* Unknown bot internal error, see server logs "
                 "for more information|.".replace('|', '\\'),
            parse_mode=ParseMode.MARKDOWN_V2
        )

    def _command_start(self, update: Update, context: CallbackContext) -> None:
        """
        Handler for `/start` command.

        It sends a presentation to the user and calls the help command.

        Args:
            update: The update to be handled.
            context: The context object for the update.
        """
        update.message.reply_text(
            text="Welcome to the *Surveillance Bot*",
            parse_mode=ParseMode.MARKDOWN_V2
        )
        self._command_help(update, context)

    def _get_reply_keyboard(
            self,
            is_active: Optional[bool] = None
    ) -> ReplyKeyboardMarkup:
        """
        Generates Reply Keyboard content.

        Args:
            is_active: Overrides surveillance mode status.

        Returns:
            ReplyKeyboardMarkup instance with the menu content.

        """
        active = self.camera.is_surveillance_active \
            if is_active is None else is_active
        custom_keyboard = [
            [
                '/get_photo',
                '/get_video'
            ],
            [
                '/surveillance_{}'.format('stop' if active else 'start')
            ]
        ]
        return ReplyKeyboardMarkup(
            custom_keyboard,
            resize_keyboard=True
        )

    def _command_help(
            self,
            update: Update,
            _: CallbackContext
    ) -> None:
        """
        Shows a help message listing all available commands.

        This command also sends the custom keyboard to the user.

        Args:
            update: The update to be handled.
        """
        update.message.reply_text(
            text="With this bot, photos or videos can be taken with the cam "
                 "upon request|. A surveillance mode is also included|. This "
                 "mode warns you when it detects movement and it will start "
                 "recording a video|. Whilst recording, photos will be taken "
                 "and sent periodically|.\n"
                 "\n"
                 "These are the available commands:\n"
                 "\n"
                 "*On Demand commands*\n"
                 "/get|_photo |- Takes a picture from the cam\n"
                 "/get|_video |- Takes a video from the cam\n"
                 "\n"
                 "*Surveillance Mode commands*\n"
                 "/surveillance|_start |- Starts surveillance mode\n"
                 "/surveillance|_stop |- Stops surveillance mode\n"
                 "/surveillance|_status |- Indicates if surveillance mode "
                 "is active or not\n"
                 "\n"
                 "*General commands*\n"
                 "/config |- Invokes configuration menu\n"
                 "/stop|_config |- Abort configuration sequence\n"
                 "/help |- Shows this help text\n"
                 "".replace('|', '\\'),
            parse_mode=ParseMode.MARKDOWN_V2,
            reply_markup=self._get_reply_keyboard()
        )

    def _command_get_photo(
            self,
            update: Update,
            context: CallbackContext
    ) -> None:
        """
        Handler for `/get_photo` command.

        It takes a single shot and sends it to the user.

        Args:
            update: The update to be handled.
            context: The context object for the update.
        """
        # Retrieves configuration
        timestamp = context.bot_data[BotConfig.TIMESTAMP]

        # Uploads photo
        context.bot.send_chat_action(
            chat_id=update.message.chat_id,
            action=ChatAction.UPLOAD_PHOTO
        )
        context.bot.send_photo(
            chat_id=update.message.chat_id,
            photo=self.camera.get_photo(timestamp=timestamp)
        )

    def _command_get_video(
            self,
            update: Update,
            context: CallbackContext
    ) -> None:
        """
        Handler for `/get_video` command.

        It takes a video and sends it to the user.

        Args:
            update: The update to be handled.
            context: The context object for the update.
        """
        # Retrieves configuration
        timestamp = context.bot_data[BotConfig.TIMESTAMP]
        seconds = context.bot_data[BotConfig.OD_VIDEO_DURATION]

        # Sends waiting message
        message = update.message.reply_text(
            text=f'Recording a {seconds} seconds video...'
        )

        # Records video
        context.bot.send_chat_action(
            chat_id=update.message.chat_id,
            action=ChatAction.RECORD_VIDEO
        )
        video = self.camera.get_video(timestamp=timestamp, seconds=seconds)

        # Uploads video
        context.bot.send_chat_action(
            chat_id=update.message.chat_id,
            action=ChatAction.UPLOAD_VIDEO
        )
        context.bot.send_video(
            chat_id=update.message.chat_id,
            video=video
        )

        # Deletes waiting message
        context.bot.delete_message(
            chat_id=update.message.chat_id,
            message_id=message.message_id
        )

    @run_async
    def _command_surveillance_start(
            self,
            update: Update,
            context: CallbackContext
    ) -> None:
        """
        Handler for `/surveillance_start` command.

        It starts the surveillance mode. In this mode the is waiting for
        motion detection, when this happens it sends a message to the user
        and start to record a video, sending pictures in regular intervals
        during the video recording. After that it goes back to the waiting
        state.

        Args:
            update: The update to be handled.
            context: The context object for the update.
        """
        # Check if surveillance is already started
        if self.camera.is_surveillance_active:
            update.message.reply_text(
                text='Error! Surveillance is already started'
            )
            self.logger.warning("Surveillance already started")
            return

        # Retrieve configuration
        timestamp = context.bot_data[BotConfig.TIMESTAMP]
        video_seconds = context.bot_data[BotConfig.SRV_VIDEO_DURATION]
        picture_interval = context.bot_data[BotConfig.SRV_PICTURE_INTERVAL]
        motion_contours = context.bot_data[BotConfig.SRV_MOTION_CONTOURS]

        # Starts surveillance
        waiting_message = None
        self.logger.info('Surveillance mode start')
        update.message.reply_text(
            text="Surveillance mode started",
            reply_markup=self._get_reply_keyboard(True)
        )
        for data in self.camera.surveillance_start(
                timestamp=timestamp,
                video_seconds=video_seconds,
                picture_seconds=picture_interval,
                contours=motion_contours
        ):
            if 'detected' in data:
                update.message.reply_text(
                    text='*MOTION DETECTED|!*'.replace('|', '\\'),
                    parse_mode=ParseMode.MARKDOWN_V2
                )
                waiting_message = update.message.reply_text(
                    text=f'Recording a {video_seconds} seconds video and '
                         f'taking {video_seconds // picture_interval} '
                         f'photos...'
                )
                context.bot.send_chat_action(
                    chat_id=update.message.chat_id,
                    action=ChatAction.RECORD_VIDEO
                )
            if 'photo' in data:
                context.bot.send_chat_action(
                    chat_id=update.message.chat_id,
                    action=ChatAction.UPLOAD_PHOTO
                )
                context.bot.send_photo(
                    chat_id=update.message.chat_id,
                    photo=data['photo'],
                    caption=f'Capture {data["id"]}/{data["total"]}'
                )
                context.bot.send_chat_action(
                    chat_id=update.message.chat_id,
                    action=ChatAction.RECORD_VIDEO
                )
            if 'video' in data:
                context.bot.send_chat_action(
                    chat_id=update.message.chat_id,
                    action=ChatAction.UPLOAD_VIDEO
                )
                context.bot.send_video(
                    chat_id=update.message.chat_id,
                    video=data['video']
                )
                if waiting_message:
                    context.bot.delete_message(
                        chat_id=update.message.chat_id,
                        message_id=waiting_message.message_id
                    )
                    waiting_message = None

        if waiting_message:
            context.bot.delete_message(
                chat_id=update.message.chat_id,
                message_id=waiting_message.message_id
            )
        update.message.reply_text(
            text="Surveillance mode stopped",
            reply_markup=self._get_reply_keyboard()
        )
        self.logger.info('Surveillance mode stop')

    def _command_surveillance_stop(
            self,
            update: Update,
            _: CallbackContext
    ) -> None:
        """
        Handler for `/surveillance_stop` command.

        This method stops the surveillance mode.

        Args:
            update: The update to be handled.
        """
        # Checks if surveillance is not running.
        if not self.camera.is_surveillance_active:
            update.message.reply_text(
                text="Error! Surveillance is not started"
            )
            self.logger.warning("Surveillance is not started")
            return

        # Stop surveillance.
        self.camera.surveillance_stop()

    def _command_surveillance_status(
            self,
            update: Update,
            _: CallbackContext
    ) -> None:
        """
        Handler for `/surveillance_stats` command.

        This method informs to the user whether surveillance mode is active
        or not.

        Args:
            update: The update to be handled.
        """
        if self.camera.is_surveillance_active:
            update.message.reply_text(text="Surveillance mode is active")
        else:
            update.message.reply_text(text="Surveillance mode is not active")



"""
Module for bot configuration functionality.

This module contains the `BotConfig` class that implements a conversational
sequence in order to configure the bot behavior.
"""
from typing import TYPE_CHECKING, Callable, List

from telegram import (
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    ParseMode,
    Update
)
from telegram.ext import (
    CallbackContext,
    CallbackQueryHandler,
    ConversationHandler,
    Filters,
    MessageHandler
)

if TYPE_CHECKING:  # pragma: no cover
    from surveillance_bot.bot import Bot  # pylint: disable=cyclic-import


class BotConfig:
    """
    Class for bot configuration process implementation.

    This class contains all constants and static methods needed to implements
    a conversational sequence with the user in order to the bot behavior will
    be configured.

    It doesn't have any instance method so it doesn't need to be instantiated.
    """
    # Configuration variables
    TIMESTAMP = 'timestamp'
    OD_VIDEO_DURATION = 'od_video_duration'
    SRV_VIDEO_DURATION = 'srv_video_duration'
    SRV_PICTURE_INTERVAL = 'srv_picture_interval'
    SRV_MOTION_CONTOURS = 'srv_motion_contours'

    # State definitions for top level conversation
    MAIN_MENU, GENERAL_CONFIG, SURVEILLANCE_CONFIG = map(chr, range(3))

    # State definitions for second level conversation
    (
        CHANGE_TIMESTAMP,
        CHANGE_OD_VIDEO_DURATION,
        CHANGE_SRV_VIDEO_DURATION,
        CHANGE_SRV_PICTURE_INTERVAL,
        CHANGE_SRV_MOTION_CONTOURS
    ) = map(chr, range(3, 8))

    # State definitions for input conversation
    BOOLEAN_INPUT, INTEGER_INPUT = map(chr, range(8, 10))

    # Shortcut for ConversationHandler.END
    END = ConversationHandler.END

    # Auxiliary constants
    CURRENT_VARIABLE, RETURN_HANDLER, ENABLE, DISABLE = map(chr, range(10, 14))

    @staticmethod
    def get_config_handler(bot: 'Bot') -> ConversationHandler:
        """
        Generates the conversation handler for whole configuration process.

        Args:
            bot: The parent `Bot` instance.

        Returns:
            The instantiated `ConversationHandler`.
        """
        main_handler = ConversationHandler(
            entry_points=[bot.command_handler('config', BotConfig._main_menu)],
            states={
                BotConfig.MAIN_MENU: [
                    CallbackQueryHandler(
                        BotConfig._general_config,
                        pattern='^' + str(BotConfig.GENERAL_CONFIG) + '$'
                    ),
                    CallbackQueryHandler(
                        BotConfig._surveillance_config,
                        pattern='^' + str(BotConfig.SURVEILLANCE_CONFIG) + '$'
                    ),
                    CallbackQueryHandler(
                        BotConfig._end,
                        pattern='^' + str(BotConfig.END) + '$'
                    )
                ],
                BotConfig.GENERAL_CONFIG: [
                    CallbackQueryHandler(
                        BotConfig._change_timestamp,
                        pattern='^'
                        + str(BotConfig.CHANGE_TIMESTAMP)
                        + '$'
                    ),
                    CallbackQueryHandler(
                        BotConfig._change_od_video_duration,
                        pattern='^'
                        + str(BotConfig.CHANGE_OD_VIDEO_DURATION)
                        + '$'
                    ),
                    CallbackQueryHandler(
                        BotConfig._main_menu,
                        pattern='^' + str(BotConfig.END) + '$'
                    )
                ],
                BotConfig.SURVEILLANCE_CONFIG: [
                    CallbackQueryHandler(
                        BotConfig._change_srv_video_duration,
                        pattern='^'
                        + str(BotConfig.CHANGE_SRV_VIDEO_DURATION)
                        + '$'
                    ),
                    CallbackQueryHandler(
                        BotConfig._change_srv_picture_interval,
                        pattern='^'
                        + str(BotConfig.CHANGE_SRV_PICTURE_INTERVAL)
                        + '$'
                    ),
                    CallbackQueryHandler(
                        BotConfig._change_motion_contours,
                        pattern='^'
                        + str(BotConfig.CHANGE_SRV_MOTION_CONTOURS)
                        + '$'
                    ),
                    CallbackQueryHandler(
                        BotConfig._main_menu,
                        pattern='^' + str(BotConfig.END) + '$'
                    )
                ],
                BotConfig.BOOLEAN_INPUT: [
                    CallbackQueryHandler(
                        BotConfig._boolean_input,
                        pattern='^'
                        + str(BotConfig.ENABLE)
                        + '$|^'
                        + str(BotConfig.DISABLE)
                        + '$'
                    )
                ],
                BotConfig.INTEGER_INPUT: [
                    MessageHandler(
                        Filters.text,
                        BotConfig._integer_input
                    )
                ]
            },
            fallbacks=[bot.command_handler('stop_config', BotConfig._end)],
        )

        return main_handler

    @staticmethod
    def ensure_defaults(context: CallbackContext) -> None:
        """
        Creates non-existent variables and populates with default values.

        Args:
            context: The context object for the update.
        """
        if BotConfig.TIMESTAMP not in context.bot_data:
            context.bot_data[BotConfig.TIMESTAMP] = True

        if BotConfig.OD_VIDEO_DURATION not in context.bot_data:
            context.bot_data[BotConfig.OD_VIDEO_DURATION] = 5

        if BotConfig.SRV_VIDEO_DURATION not in context.bot_data:
            context.bot_data[BotConfig.SRV_VIDEO_DURATION] = 30

        if BotConfig.SRV_PICTURE_INTERVAL not in context.bot_data:
            context.bot_data[BotConfig.SRV_PICTURE_INTERVAL] = 5

        if BotConfig.SRV_MOTION_CONTOURS not in context.bot_data:
            context.bot_data[BotConfig.SRV_MOTION_CONTOURS] = True

    # Menus

    @staticmethod
    def _main_menu(update: Update, _: CallbackContext) -> str:
        """
        Creates the main menu and send it to the user.

        This menu links to the general configuration and to the surveillance
        mode configuration.

        Args:
            update: The update to be handled.

        Returns:
            The state MAIN_MENU.
        """
        text = "*Surveillance Telegram Bot Configuration*\n" \
               "\n" \
               "Here you can modify some bot behavior parameters|. \n" \
               "\n" \
               "While the mode is running it does not allow you any change " \
               "unless you restart it|.\n" \
               "\n" \
               "To abort type /stop|_config|.\n" \
               "\n" \
               "Select section:".replace('|', '\\')
        buttons = [
            [InlineKeyboardButton(
                text='General configuration',
                callback_data=str(BotConfig.GENERAL_CONFIG)
            )],
            [InlineKeyboardButton(
                text='Surveillance mode configuration',
                callback_data=str(BotConfig.SURVEILLANCE_CONFIG)
            )],
            [InlineKeyboardButton(
                text='Done',
                callback_data=str(BotConfig.END)
            )]
        ]

        BotConfig._render_menu(update, text, buttons)

        return BotConfig.MAIN_MENU

    @staticmethod
    def _general_config(update: Update, context: CallbackContext) -> str:
        """
        Creates the menu for general configuration and send it to the user.

        Args:
            update: The update to be handled.
            context: The context object for the update.

        Returns:
            The state GENERAL_CONFIG.
        """
        timestamp = context.bot_data[BotConfig.TIMESTAMP]
        video_duration = context.bot_data[BotConfig.OD_VIDEO_DURATION]

        timestamp_str = 'Enabled' if timestamp else 'Disabled'

        text = f"*General configuration*\n" \
               f"\n" \
               f"__Timestamp__:\n" \
               f" |- _Description_: Print a timestamp on every photo or" \
               f" video taken|.\n" \
               f" |- _Current value_: *{timestamp_str}*\n" \
               f"\n" \
               f"__On Demand video duration__:\n" \
               f" |- _Description_: Duration of the video taken with " \
               f"/get|_video command|.\n" \
               f" |- _Current value_: *{video_duration} seconds*" \
               f"".replace('|', '\\')
        buttons = [
            [InlineKeyboardButton(
                text='Timestamp',
                callback_data=str(BotConfig.CHANGE_TIMESTAMP)
            )],
            [InlineKeyboardButton(
                text='On Demand video duration',
                callback_data=str(BotConfig.CHANGE_OD_VIDEO_DURATION)
            )],
            [InlineKeyboardButton(
                text='Back',
                callback_data=str(BotConfig.END)
            )]
        ]

        BotConfig._render_menu(update, text, buttons)

        return BotConfig.GENERAL_CONFIG

    @staticmethod
    def _surveillance_config(update: Update, context: CallbackContext) -> str:
        """
        Creates the menu for the surveillance mode configuration and send it
        to the user.

        Args:
            update: The update to be handled.
            context: The context object for the update.

        Returns:
            The state SURVEILLANCE_CONFIG.
        """
        video_duration = context.bot_data[BotConfig.SRV_VIDEO_DURATION]
        picture_interval = context.bot_data[BotConfig.SRV_PICTURE_INTERVAL]
        motion_contours = context.bot_data[BotConfig.SRV_MOTION_CONTOURS]

        motion_contours_str = 'Enabled' if motion_contours else 'Disabled'

        text = f"*Surveillance Mode configuration*\n" \
               f"\n" \
               f"__Video duration__:\n" \
               f" |- _Description_: Duration of the video taken when motion " \
               f"is detected|.\n" \
               f" |- _Current value_: *{video_duration} seconds*\n" \
               f"\n" \
               f"__Picture Interval__:\n" \
               f" |- _Description_: Interval between photos taken after " \
               f"motion is detected|.\n" \
               f" |- _Current value_: *{picture_interval} seconds*\n" \
               f"\n" \
               f"__Draw motion contours__:\n" \
               f" |- _Description_: Draws a rectangle around the objects in " \
               f"motion|.\n" \
               f" |- _Current value_: *{motion_contours_str}*" \
               f"".replace('|', '\\')
        buttons = [
            [InlineKeyboardButton(
                text='Video duration',
                callback_data=str(BotConfig.CHANGE_SRV_VIDEO_DURATION)
            )],
            [InlineKeyboardButton(
                text='Picture Interval',
                callback_data=str(BotConfig.CHANGE_SRV_PICTURE_INTERVAL)
            )],
            [InlineKeyboardButton(
                text='Draw motion contours',
                callback_data=str(BotConfig.CHANGE_SRV_MOTION_CONTOURS)
            )],
            [InlineKeyboardButton(
                text='Back',
                callback_data=str(BotConfig.END)
            )]
        ]

        BotConfig._render_menu(update, text, buttons)

        return BotConfig.SURVEILLANCE_CONFIG

    @staticmethod
    def _render_menu(
            update: Update,
            text: str,
            buttons: List[List[InlineKeyboardButton]]
    ) -> None:
        """
        Builds the inline keyboard for the menu and sends all to the user.

        Args:
            update: The update to be handled.
            text: Text for the menu caption.
            buttons: Array of button rows,
                each represented by an Array of InlineKeyboardButton objects.
        """
        keyboard = InlineKeyboardMarkup(buttons)

        if update.message:
            update.message.reply_text(
                text=text,
                reply_markup=keyboard,
                parse_mode=ParseMode.MARKDOWN_V2
            )
        else:
            update.callback_query.answer()
            update.callback_query.edit_message_text(
                text=text,
                reply_markup=keyboard,
                parse_mode=ParseMode.MARKDOWN_V2
            )

    # General configuration options.

    @staticmethod
    def _change_timestamp(
            update: Update,
            context: CallbackContext
    ) -> str:
        """
        Prepares all required data to request the TIMESTAMP configuration to
        the user.

        Args:
            update: The update to be handled.
            context: The context object for the update.

        Returns:
            The state BOOLEAN_INPUT through `_boolean_question` method.
        """
        timestamp = context.bot_data[BotConfig.TIMESTAMP]

        timestamp_str = 'Enabled' if timestamp else 'Disabled'

        text = f'*Timestamp*\n' \
               f'\n' \
               f'Current state: *{timestamp_str}*\n' \
               f'\n' \
               f'Select state for time stamping:'

        return BotConfig._boolean_question(
            update,
            context,
            text,
            BotConfig.TIMESTAMP,
            BotConfig._general_config
        )

    @staticmethod
    def _change_od_video_duration(
            update: Update,
            context: CallbackContext
    ) -> str:
        """
        Prepares all required data to request the OD_VIDEO_DURATION
        configuration to the user.

        Args:
            update: The update to be handled.
            context: The context object for the update.

        Returns:
            The state INTEGER_INPUT through `_integer_question` method.
        """
        video_duration = context.bot_data[BotConfig.OD_VIDEO_DURATION]

        text = f'*On Demand video duration*\n' \
               f'\n' \
               f'Current value: *{video_duration}*\n' \
               f'\n' \
               f'Type value for video duration:'

        return BotConfig._integer_question(
            update,
            context,
            text,
            BotConfig.OD_VIDEO_DURATION,
            BotConfig._general_config
        )

    # Surveillance mode configuration options.

    @staticmethod
    def _change_srv_video_duration(
            update: Update,
            context: CallbackContext
    ) -> str:
        """
        Prepares all required data to request the SRV_VIDEO_DURATION
        configuration to the user.

        Args:
            update: The update to be handled.
            context: The context object for the update.

        Returns:
            The state INTEGER_INPUT through `_integer_question` method.
        """
        video_duration = context.bot_data[BotConfig.SRV_VIDEO_DURATION]

        text = f'*Surveillance video duration*\n' \
               f'\n' \
               f'Current value: *{video_duration}*\n' \
               f'\n' \
               f'Type value for video duration:'

        return BotConfig._integer_question(
            update,
            context,
            text,
            BotConfig.SRV_VIDEO_DURATION,
            BotConfig._surveillance_config
        )

    @staticmethod
    def _change_srv_picture_interval(
            update: Update,
            context: CallbackContext
    ) -> str:
        """
        Prepares all required data to request the SRV_PICTURE_INTERVAL
        configuration to the user.

        Args:
            update: The update to be handled.
            context: The context object for the update.

        Returns:
            The state INTEGER_INPUT through `_integer_question` method.
        """
        picture_interval = context.bot_data[BotConfig.SRV_PICTURE_INTERVAL]

        text = f'*Surveillance picture interval*\n' \
               f'\n' \
               f'Current value: *{picture_interval}*\n' \
               f'\n' \
               f'Type value for picture interval:'

        return BotConfig._integer_question(
            update,
            context,
            text,
            BotConfig.SRV_PICTURE_INTERVAL,
            BotConfig._surveillance_config
        )

    @staticmethod
    def _change_motion_contours(
            update: Update,
            context: CallbackContext
    ) -> str:
        """
        Prepares all required data to request the SRV_MOTION_CONTOURS
        configuration to the user.

        Args:
            update: The update to be handled.
            context: The context object for the update.

        Returns:
            The state BOOLEAN_INPUT through `_boolean_question` method.
        """
        motion_contours = context.bot_data[BotConfig.SRV_MOTION_CONTOURS]

        motion_contours_str = 'Enabled' if motion_contours else 'Disabled'

        text = f'*Motion contours*\n' \
               f'\n' \
               f'Current state: *{motion_contours_str}*\n' \
               f'\n' \
               f'Select state for motion contours:'

        return BotConfig._boolean_question(
            update,
            context,
            text,
            BotConfig.SRV_MOTION_CONTOURS,
            BotConfig._surveillance_config
        )

    # Questions helpers.

    @staticmethod
    def _boolean_question(
            update: Update,
            context: CallbackContext,
            text: str,
            current_variable: str,
            return_handler: Callable[[Update, CallbackContext], str]
    ) -> str:
        """
        Builds a boolean question to send it to the user using received data.

        Args:
            update: The update to be handled.
            context: The context object for the update.
            text: Message to be shown to the users.
            current_variable: Variable to be set.
            return_handler: Handler to be called with the user response.

        Returns:
            The state BOOLEAN_INPUT.
        """
        context.user_data[BotConfig.CURRENT_VARIABLE] = current_variable
        context.user_data[BotConfig.RETURN_HANDLER] = return_handler

        buttons = [[
            InlineKeyboardButton(
                text='Enable',
                callback_data=str(BotConfig.ENABLE)
            ),
            InlineKeyboardButton(
                text='Disable',
                callback_data=str(BotConfig.DISABLE)
            ),
        ]]
        keyboard = InlineKeyboardMarkup(buttons)

        update.callback_query.answer()
        update.callback_query.edit_message_text(
            text=text,
            reply_markup=keyboard,
            parse_mode=ParseMode.MARKDOWN_V2
        )

        return BotConfig.BOOLEAN_INPUT

    @staticmethod
    def _integer_question(
            update: Update,
            context: CallbackContext,
            text: str,
            current_variable: str,
            return_handler: Callable[[Update, CallbackContext], str]
    ) -> str:
        """
        Builds a integer question to send it to the user using received data.

        Args:
            update: The update to be handled.
            context: The context object for the update.
            text: Message to be shown to the users.
            current_variable: Variable to be set.
            return_handler: Handler to be called with the user response.

        Returns:
            The state INTEGER_INPUT.
        """
        context.user_data[BotConfig.CURRENT_VARIABLE] = current_variable
        context.user_data[BotConfig.RETURN_HANDLER] = return_handler

        update.callback_query.answer()
        update.callback_query.edit_message_text(
            text=text,
            parse_mode=ParseMode.MARKDOWN_V2
        )

        return BotConfig.INTEGER_INPUT

    # Input handlers.

    @staticmethod
    def _boolean_input(update: Update, context: CallbackContext) -> str:
        """
        Receive a boolean input from the user and saves the value into
        corresponding variable.

        Args:
            update: The update to be handled.
            context: The context object for the update.

        Returns:
            The execution of the previously stored handler.
        """
        context.bot_data[
            context.user_data[BotConfig.CURRENT_VARIABLE]
        ] = update.callback_query.data == BotConfig.ENABLE

        return context.user_data[BotConfig.RETURN_HANDLER](update, context)

    @staticmethod
    def _integer_input(update: Update, context: CallbackContext) -> str:
        """
        Receive a integer input from the user, validates it, and saves the
        value into corresponding variable.

        Args:
            update: The update to be handled.
            context: The context object for the update.

        Returns:
            The execution of the previously stored handler or the state
                INTEGER_INPUT in case of validation error.
        """
        try:
            value = int(update.message.text)
            assert value > 0
            assert value < 100
        except (ValueError, AssertionError):
            update.message.reply_text(
                text='Invalid value, insert an integer number between 1 and 99'
            )
            return BotConfig.INTEGER_INPUT

        context.bot_data[
            context.user_data[BotConfig.CURRENT_VARIABLE]
        ] = value

        return context.user_data[BotConfig.RETURN_HANDLER](update, context)

    @staticmethod
    def _end(update: Update, context: CallbackContext) -> int:
        """
        Handler to end the configuration sequence.

        Args:
            update: The update to be handled.
            context: The context object for the update.

        Returns:
            The state END.
        """
        context.user_data.clear()

        if update.callback_query:
            update.callback_query.answer()
            update.callback_query.edit_message_text(text='Configuration done.')
        else:
            update.message.reply_text(text='Configuration canceled.')

        return BotConfig.END



"""
Module for camera related functionality.

This module implements both camera management and a top layer for related
operations such as motion detection or output file generation. All
camera and image operations are performed using OpenCV.
"""
import os
from datetime import datetime
from io import BytesIO
from tempfile import TemporaryDirectory
from threading import Lock, Thread
from time import time
from typing import IO, Any, Dict, Iterator, List, Optional, Tuple

import cv2
import numpy as np


# Exceptions
class CameraError(Exception):
    """Base class for Camera related errors."""


class CameraConnectionError(CameraError):
    """Raised when connection with the camera fails."""


class CodecNotAvailable(CameraError):
    """Raised when a suitable video codec is not available."""


# Classes
class CameraDevice:
    """
    Class for camera hardware handling.

    This class handles camera hardware using threading to perform reading
    operations on the camera.

    Args:
        cam_id: ID of the video capturing device to open.
    """
    def __init__(self, cam_id=0) -> None:
        self._device = cv2.VideoCapture(cam_id)
        if not self._device.isOpened():
            raise CameraConnectionError
        self._frame = self._device.read()[1]

        self._frame_count = 0
        self._start_time = 0.0

        self._running = False
        self._thread: Optional[Thread] = None
        self._lock = Lock()

    def start(self) -> None:
        """Starts frame grabbing process."""
        if not self._running:
            self._frame_count = 0
            self._start_time = time()
            self._running = True
            self._thread = Thread(target=self._update, daemon=True)
            self._thread.start()

    def _update(self) -> None:
        """
        Updates data with latest frame available.

        This functions is executed in a separated thread because frame
        reading is a blocking operation. Every frame grabbed is stored
        temporarily.
        """
        while self._running:
            frame = self._device.read()[1]
            with self._lock:
                self._frame = frame
                self._frame_count += 1

    def read(self, timestamp=True) -> Tuple[int, np.ndarray]:
        """
        Returns the latest stored frame.

        Args:
            timestamp: If True a timestamp is printed on the returned frame.

        Returns:
            A tuple with the unique identifier of the frame and the frame
            itself.
        """
        with self._lock:
            frame_id = self._frame_count
            frame = self._frame.copy()
        if timestamp:
            self._add_timestamp(frame)
        return frame_id, frame

    def stop(self) -> None:
        """Stops frame grabbing process."""
        self._running = False
        if self._thread:
            self._thread.join()

    @property
    def fps(self) -> float:
        """
        Calculates actual frames per seconds value.

        Note:
            This is necessary because FPS value returned by OpenCV could be
            not accurate.

        Returns:
            Current FPS value.
        """
        return self._frame_count / (time() - self._start_time)

    @property
    def frame_size(self) -> Tuple[int, int]:
        """
        Gets the frame resolution based in the last grabbed frame.

        Returns:
            Tuple with frame width and height.
        """
        height, width, _ = self._frame.shape
        return width, height

    @staticmethod
    def _add_timestamp(frame: np.ndarray) -> None:
        """Prints timestamp on the given frame."""
        now = str(datetime.now())[:-7]
        org = (1, frame.shape[0] - 3)
        font = cv2.FONT_HERSHEY_PLAIN
        size = 1
        cv2.putText(frame, now, org, font, size, (0, 0, 0), 2)
        cv2.putText(frame, now, org, font, size, (255, 255, 255), 1)

    def __del__(self) -> None:
        """Releases video capture before object is destroyed."""
        self._device.release()


class Camera:
    """
    Top level class for camera operations performing.

    This class provide high level operations returning images and videos in
    a file object.

    Args:
        cam_id: ID of the video capturing device to open.
    """

    STATE_IDLE = 'idle'
    """Awaiting for motion detection."""

    STATE_MOTION_DETECTED = 'motion_detected'
    """After motion have been detected."""

    def __init__(self, cam_id=0) -> None:
        self._camera = CameraDevice(cam_id)
        self._surveillance_mode = False
        self._tempdir = TemporaryDirectory()
        self._codec = self.get_supported_codec()
        if not self._codec:
            raise CodecNotAvailable

    def start(self) -> None:
        """ Starts camera device."""
        self._camera.start()

    def stop(self) -> None:
        """Stops camera device."""
        self._camera.stop()

    def get_photo(self, timestamp=True) -> IO:
        """
        Takes a single shot.

        Args:
            timestamp: Adds time stamping on the photo.

        Returns:
            File object with the photo taken.
        """
        frame = self._camera.read(timestamp=timestamp)[1]
        return BytesIO(cv2.imencode(".jpg", frame)[1])

    def get_video(self, timestamp=True, seconds=5) -> IO:
        """Takes a video.

        Args:
            timestamp: Adds time stamping on the video.
            seconds: Video duration.

        Returns:
            File object with the video taken.
        """
        path, video_writer = self._create_video_file('on_demand')
        n_frames = self._camera.fps * seconds

        processed: List[int] = []
        while len(processed) < n_frames:
            frame_id, frame = self._camera.read(timestamp=timestamp)
            if frame_id not in processed:
                processed.append(frame_id)
                video_writer.write(frame)

        video_writer.release()
        return open(path, 'rb')

    @staticmethod
    def _get_motion_contours(
            frame_1: np.ndarray,
            frame_2: np.ndarray
    ) -> List[np.ndarray]:
        """
        Detects motion and find the contours for every motion detected.

        This method receives two consecutive frames and detects changes
        between them.

        Args:
            frame_1: First of the two consecutive frames.
            frame_2: Second of the two consecutive frames.

        Returns:
            A list with the contours found.
        """
        frame_delta = cv2.absdiff(frame_1, frame_2)
        thresh = cv2.threshold(frame_delta, 5, 255, cv2.THRESH_BINARY)[1]

        kernel = np.ones((40, 40), np.uint8)
        thresh = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel)
        return cv2.findContours(
            thresh.copy(),
            cv2.RETR_EXTERNAL,
            cv2.CHAIN_APPROX_SIMPLE
        )[-2]

    @staticmethod
    def _draw_contours(frame: np.ndarray, contour: np.ndarray) -> None:
        """
        Draws a rectangle on the frame that marks a contour.

        This method gets a bounding rectangle for a given contour and
        draws this rectangle on the frame.

        Args:
            frame: Frame on which to draw the rectangles.
            contour: Contour to be marked.
        """
        (x, y, width, height) = cv2.boundingRect(contour)
        cv2.rectangle(frame, (x, y), (x + width, y + height), (0, 255, 0), 1)

    def _motion_detection(
            self,
            timestamp=True,
            contours=True
    ) -> Iterator[Tuple[bool, int, np.ndarray]]:
        """
        Executes motion detection operation during surveillance mode.

        This generator grabs frames continuously and search for motion
        yielding every frame grabbed even motion is detected or not. In
        frames with motion detected the contour is drawn on it.

        Args:
            timestamp: Adds time stamping on the frames.
            contours: Draws motion contours on the frames.

        Yields:
            A tuple with three values.
                * True if motion is detected or False if not.
                * The unique identifier of the frame.
                * The frame itself.
        """
        self._surveillance_mode = True
        previous_frame = None
        last_frame_id = 0

        while self._surveillance_mode:
            frame_id, frame = self._camera.read(timestamp=timestamp)
            if frame_id == last_frame_id:
                continue
            last_frame_id = frame_id

            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            gray = cv2.GaussianBlur(gray, (21, 21), 0)

            if previous_frame is None:
                previous_frame = gray
                continue

            motion_contours = self._get_motion_contours(previous_frame, gray)

            detected = False
            for contour in motion_contours:
                if cv2.contourArea(contour) < 2000:
                    continue
                if contours:
                    self._draw_contours(frame, contour)
                detected = True

            previous_frame = gray
            yield detected, frame_id, frame

    def surveillance_start(
            self,
            timestamp=True,
            video_seconds=30,
            picture_seconds=5,
            contours=True
    ) -> Iterator[Dict[str, Any]]:
        """
        Starts surveillance mode, waiting for motion detection.

        In this mode it waits until motion is detected, when this happens a
        message is yielded and it starts to record a video, yielding single
        photos during this process until video is yielded. After that it
        goes back to the initial state.

        Args:
            timestamp: Adds time stamping on the frames.
            video_seconds: Video duration.
            picture_seconds: Time elapsed between pictures.
            contours: Draws motion contours on the frames.

        Yields:
            A dict with three possible configurations
                * ``{'detected': True}``
                * ``{'video': <IO>}``
                * ``{'photo': <IO>, 'id': <int>, 'total': <int>}``
        """
        status = Camera.STATE_IDLE
        fps = self._camera.fps
        processed = []
        n_frames = 0
        path = ''
        video_writer: cv2.VideoWriter = cv2.VideoWriter()

        for detected, frame_id, frame in self._motion_detection(
                timestamp=timestamp,
                contours=contours
        ):
            if status == Camera.STATE_IDLE:
                if detected:
                    yield {'detected': True}
                    status = Camera.STATE_MOTION_DETECTED
                    path, video_writer = self._create_video_file('on_motion')
                    n_frames = fps * video_seconds
                    processed = [frame_id]
                    video_writer.write(frame)
            if status == Camera.STATE_MOTION_DETECTED:
                if not len(processed) % int(fps * picture_seconds):
                    yield {
                        'photo': BytesIO(cv2.imencode(".jpg", frame)[1]),
                        'id': (len(processed) // int(fps * picture_seconds)),
                        'total': video_seconds // picture_seconds
                    }
                if len(processed) < n_frames:
                    if frame_id not in processed:
                        processed.append(frame_id)
                        video_writer.write(frame)
                else:
                    video_writer.release()
                    yield {'video': open(path, 'rb')}
                    status = Camera.STATE_IDLE

    def surveillance_stop(self) -> None:
        """Stops surveillance mode."""
        self._surveillance_mode = False

    @property
    def is_surveillance_active(self) -> bool:
        """Return if surveillance mode is active or not."""
        return self._surveillance_mode

    def _create_video_file(
            self,
            event_type: str
    ) -> Tuple[str, cv2.VideoWriter]:
        """
        Initializes a temporary video file and a video writer.


        Args:
            event_type: Event type string for filename generation.

        Returns:
            A tuple with the file path and the OpenCV video writer object.
        """
        now_str = str(datetime.now())[:-7]
        table = str.maketrans(': ', '-_')
        filename = now_str.translate(table) + f'_{event_type}.mp4'

        path = os.path.join(self._tempdir.name, filename)
        writer = cv2.VideoWriter(
            path,
            cv2.VideoWriter_fourcc(*self._codec),
            self._camera.fps,
            self._camera.frame_size
        )
        return path, writer

    def get_supported_codec(self) -> Optional[str]:
        """
        Get the best codec supported by the encoding backend.

        This method creates a test file to check the codecs availability.

        Returns:
            The fourCC string for the codec or None if there are no available
                codec.
        """
        path = os.path.join(self._tempdir.name, 'test_codec.mp4')
        for codec in ['avc1', 'mp4v']:
            writer = cv2.VideoWriter(
                path,
                cv2.VideoWriter_fourcc(*codec),
                1,
                (2, 2)
            )
            if writer.isOpened():
                writer.release()
                return codec
        return None



import setuptools


setuptools.setup(
    name="surveillance_bot",
    version="1.0",
    packages=['surveillance_bot'],
    install_requires=["python-telegram-bot>=12", "opencv-python>=4"],
    python_requires='>=3.6, <3.9',
    entry_points={
        'console_scripts': [
            'surveillance_bot = surveillance_bot.main:main'
        ]
    },
    author="Pablo Chinea",
    author_email="khertz@gmail.com",
    description="Basic video surveillance system controlled "
                "through a telegram bot.",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    license="GPL v3",
    url="https://github.com/pchinea/telegram-surveillance-bot",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Communications :: Chat",
        "Topic :: Internet",
        "Topic :: Multimedia :: Graphics :: Capture :: Digital Camera",
        "Topic :: Multimedia :: Video :: Capture",
        "Topic :: Security",
        "Topic :: Utilities"
    ]

)



"""Main script for Surveillance Telegram Bot."""
import logging
import os

from surveillance_bot import bot

# Configuration environment variables.
AUTHORIZED_USER = os.environ.get('AUTHORIZED_USER', '')
BOT_API_TOKEN = os.environ.get('BOT_API_TOKEN', '')
PERSISTENCE_DIR = os.environ.get('PERSISTENCE_DIR', None)
LOG_LEVEL = os.environ.get('LOG_LEVEL', logging.WARNING)
BOT_LOG_LEVEL = os.environ.get('BOT_LOG_LEVEL', None)


# Logging config.
logging.basicConfig(
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    level=LOG_LEVEL
)
logging.captureWarnings(True)
logging.getLogger('py.warnings').setLevel(logging.ERROR)


def main() -> None:
    """Surveillance Telegram Bot start up function."""
    surveillance_bot = bot.Bot(
        token=BOT_API_TOKEN,
        username=AUTHORIZED_USER,
        persistence_dir=PERSISTENCE_DIR,
        log_level=BOT_LOG_LEVEL
    )
    surveillance_bot.start()


if __name__ == '__main__':  # pragma: no cover
    main()
